"""Fields with customizable lookup fields."""

import base64
import json

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers
from rest_framework.exceptions import ParseError


__all__ = [
    'CustomizableLookupRelatedField',
    'CustomizableLookupManyRelatedField',
]


class SetupRequiredAttribute(object):

    def __init__(self, name, start_value=None):
        self._name = '_store_{}'.format(name)
        self._start_value = start_value

    def __get__(self, instance, owner):
        """Perform CustomizableLookupRelatedField setup if it's not already done."""
        if not instance:
            return self

        if not instance._setup_done:  # noqa
            try:
                instance.setup_field_lookup()
                instance._setup_exception = None
            except Exception as e:
                instance._setup_exception = e

        if instance._setup_exception is not None:
            raise instance._setup_exception

        return getattr(instance, self._name, self._start_value)

    def __set__(self, instance, value):
        if not instance:
            return

        setattr(instance, self._name, value)


class CustomizableLookupRelatedField(serializers.RelatedField):

    """RelatedField using supplied in header field for lookup and read value.
    Default lookup field is integer pk.
    """

    default_error_messages = {
        'required': _("This field is required."),
        'does_not_exist': _("Invalid {lookup_field} \"{value}\" - object does not exist."),
        'incorrect_type': _("Incorrect type. Expected {lookup_field} value, received {data_type}."),
        'not_allowed_field': _(
            "Error in content lookup fields header. "
            "Field {lookup_field} is not allowed lookup field for {field_name}. Only {allowed_fields} are allowed."
        ),
        'invalid_header_data': _("Invalid content of content lookup fields header!"),
    }

    LOOKUP_FIELDS_FIELD_KEY = 'field'
    LOOKUP_FIELDS_FILTER_LOOKUP_SUFFIX_KEY = 'filter_lookup_suffix'
    LOOKUP_FIELDS_AUTO = '__auto__'
    LOOKUP_FIELDS_AUTO_TO_REPRESENTATION_ATTR = 'pk'

    lookup_field_name = SetupRequiredAttribute('lookup_field_name')
    lookup_field = SetupRequiredAttribute('lookup_field')
    filter_lookup_suffix = SetupRequiredAttribute('filter_lookup_suffix')
    full_field_path = SetupRequiredAttribute('full_field_path')
    root_parent = SetupRequiredAttribute('root_parent')

    def __init__(self, **kwargs):
        self._setup_done = False
        self._setup_exception = None
        self.no_pk_lookup = kwargs.pop('no_pk_lookup', False)
        self.lookup_fields = self.get_default_lookup_fields(override=kwargs.pop('lookup_fields', {}))
        self.default_lookup_field_name = kwargs.pop('default_lookup_field_name', 'pk')
        self.http_header_name = kwargs.pop(
            'http_header_name',
            getattr(settings, 'LOOKUP_FIELDS_CONTENT_LOOKUP_FIELDS_HEADER_DEFAULT_NAME', 'HTTP_CONTENT_LOOKUP_FIELDS'),
        )
        super(CustomizableLookupRelatedField, self).__init__(**kwargs)

    def use_pk_only_optimization(self):
        return self.lookup_field_name == 'pk'

    def get_instance(self, value):
        """Method for easier instance getting overriding. Should return instance or raise proper DoesNotExist."""
        return self.get_queryset().get(**self.get_instance_filter_kwargs(value))

    def to_internal_value(self, data):
        data = self.lookup_field.to_internal_value(data)
        try:
            return self.get_instance(data)
        except ObjectDoesNotExist:
            self.fail(
                'does_not_exist',
                lookup_field=self.lookup_field_name,
                value=data,
            )
        except (TypeError, ValueError):
            self.fail(
                'incorrect_type',
                lookup_field=self.lookup_field_name,
                data_type=type(data).__name__,
            )

    def to_representation(self, value):
        value = getattr(
            value,
            self.LOOKUP_FIELDS_AUTO_TO_REPRESENTATION_ATTR if self.lookup_field_name == self.LOOKUP_FIELDS_AUTO else self.lookup_field_name,
        )
        if value is not None:
            return self.lookup_field.to_representation(value)
        return None

    def setup_field_lookup(self):
        if self.parent and self.parent.parent is not None or self._is_root_parent(self.parent):
            self._setup_done = True
            full_field_path = [] if not self.field_name else [self.field_name]
            parent = self.parent
            self._append_if_not_empty(full_field_path, parent.field_name)

            while parent.parent:
                parent = parent.parent
                self._append_if_not_empty(full_field_path, parent.field_name)

            self.full_field_path = '.'.join(reversed(full_field_path))
            self.root_parent = parent

            self.lookup_field_name = self.get_lookup_field_name()
            if isinstance(self.lookup_fields[self.lookup_field_name], dict):
                self.lookup_field = self.lookup_fields[self.lookup_field_name][self.LOOKUP_FIELDS_FIELD_KEY]
                self.filter_lookup_suffix = self.lookup_fields[self.lookup_field_name].get(
                    self.LOOKUP_FIELDS_FILTER_LOOKUP_SUFFIX_KEY,
                    None,
                )
            else:
                self.lookup_field = self.lookup_fields[self.lookup_field_name]

    def validate_empty_values(self, data):
        if self.lookup_field_name == self.LOOKUP_FIELDS_AUTO:
            return True, None
        return super().validate_empty_values(data)

    ##########
    # Helpers
    ##########

    def get_lookup_field_name(self):
        lookup_field_name = self.get_data_from_request_header().get(self.full_field_path, self.default_lookup_field_name)
        if lookup_field_name not in self.lookup_fields:
            self.fail(
                'not_allowed_field',
                lookup_field=lookup_field_name,
                field_name=self.full_field_path,
                allowed_fields=', '.join(self.lookup_fields.keys()),
            )
        return lookup_field_name

    def get_data_from_request_header(self):
        request = self.get_request()
        if request and self.http_header_name in request.META:
            try:
                data = json.loads(base64.b64decode(request.META[self.http_header_name]).decode('utf-8'))
                if isinstance(data, dict):
                    return data
                raise TypeError()
            except (ValueError, TypeError):
                raise ParseError(self.default_error_messages['invalid_header_data'])
        return {}

    def get_request(self):
        """Get request from context of root parent."""
        if self.root_parent and hasattr(self.root_parent, '_context'):
            return self.root_parent._context.get('request')  # noqa

    def get_default_lookup_fields(self, override=None):
        fields = {
            'pk': serializers.IntegerField(),
        } if not self.no_pk_lookup else {}
        if override:
            fields.update(override)
        return fields

    def get_instance_filter_kwargs(self, value):
        lookup_suffix = self.filter_lookup_suffix or ''
        return {
            '{}{}'.format(self.lookup_field_name, lookup_suffix): value
        }

    @staticmethod
    def _append_if_not_empty(collection, element):
        if element:
            collection.append(element)

    @staticmethod
    def _is_root_parent(parent):
        if parent and not parent.parent and getattr(parent, '_context', None):
            return True
        return False


class CustomizableLookupManyRelatedField(serializers.ManyRelatedField):

    """Left for backwards compatibility. Deprecated"""



"""Serializers with customizable lookup fields."""

from rest_framework import serializers

from .fields import CustomizableLookupRelatedField


class CustomizableLookupFieldsModelSerializer(serializers.ModelSerializer):

    """ModelSerializer with default related field CustomizableLookupRelatedField."""

    serializer_related_field = CustomizableLookupRelatedField


class CustomizableLookupFieldsListSerializer(serializers.ListSerializer):

    """Left for backwards compatibility. Deprecated"""



"""Permission utils for customizable lookup fields."""

from typing import (
    List,
    Optional,
)


class WithPermissions(str):

    """Customizable lookup field with permission"""

    permissions: List[str]

    def __new__(cls, *args, permissions: Optional[List[str]] = None, **kwargs):
        new = super().__new__(cls, *args, **kwargs)
        new.permissions = permissions or []
        return new

    def has_permission(self, user):
        """Check is user is authenticated and has needed permissions."""
        if not user.is_authenticated:
            return False
        if not user.has_perms(self.permissions):
            return False
        return True



"""View mixins with customizable lookup field."""

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
)

from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from rest_framework.exceptions import ParseError
from rest_framework.generics import (
    GenericAPIView,
    get_object_or_404,
)

from .permissions import WithPermissions


class CustomizableLookupFieldMixin(object):

    """A mixin for API views with customizable lookup field specified in the request header Lookup-Field option"""

    DEFAULT_LOOKUP_FIELD: str = GenericAPIView.lookup_field
    HEADER_LOOKUP_FIELD_NAME: str = 'HTTP_LOOKUP_FIELD'  # matches the Lookup-Field header
    LOOKUP_FIELD_MAPPERS: Dict[str, List[Callable]] = {}
    LOOKUP_FIELD_VALIDATORS: Dict[str, List[Callable]] = {}  # Dict of lists of validators for each lookup field - field name as key
    lookup_url_kwarg = 'identifier'

    @property
    def ALLOWED_LOOKUP_FIELDS(self) -> List[str]:
        return [self.DEFAULT_LOOKUP_FIELD]

    @property
    def lookup_field(self) -> str:
        try:
            lookup_field = self.request.META[self.HEADER_LOOKUP_FIELD_NAME]
        except KeyError:
            lookup_field = self.DEFAULT_LOOKUP_FIELD

        permitted_lookup_fields = list(self._get_permitted_lookup_fields())

        if lookup_field not in permitted_lookup_fields:
            raise ParseError(_('Only {allowed_fields} lookup fields are allowed.').format(
                allowed_fields=', '.join(permitted_lookup_fields)
            ))

        validators = self.LOOKUP_FIELD_VALIDATORS.get(lookup_field, [])
        for validator in validators:
            try:
                validator(self.kwargs[self.lookup_url_kwarg])
            except ValidationError as e:
                raise ParseError('{lookup}: {message}'.format(lookup=lookup_field, message=u' '.join(e.messages)))
        return lookup_field

    def get_object(self):
        """
        Return the object the view is displaying.

        You may want to override this if you need to provide non-standard
        queryset lookups.  Eg if objects are referenced using multiple
        keyword arguments in the url conf.

        Note: due to need to inject decrypted identifier and hard to extend original function this override original one.
        """
        queryset = self.filter_queryset(self.get_queryset())

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            (self.__class__.__name__, lookup_url_kwarg)
        )
        lookup_field = self.lookup_field
        identifier = self.kwargs[lookup_url_kwarg]
        filter_kwargs = self.get_lookup_filters(lookup_field=lookup_field, identifier=identifier)
        obj = get_object_or_404(queryset, **filter_kwargs)

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    def get_lookup_filters(self, lookup_field, identifier) -> Dict[str, Any]:
        if hasattr(self, 'LOOKUP_FIELD_MAPPERS') and lookup_field in self.LOOKUP_FIELD_MAPPERS:
            lookup_field, identifier = self.LOOKUP_FIELD_MAPPERS[lookup_field](identifier)

        return {lookup_field: identifier}

    def _get_permitted_lookup_fields(self) -> Generator[str, None, None]:
        for lookup_field in self.ALLOWED_LOOKUP_FIELDS:
            if isinstance(lookup_field, WithPermissions) and not lookup_field.has_permission(self.request.user):
                continue

            yield lookup_field




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lookup_fields']

package_data = \
{'': ['*'],
 'lookup_fields': ['locale/en/LC_MESSAGES/*', 'locale/pl/LC_MESSAGES/*']}

install_requires = \
['Django>=2.0,<4.0', 'djangorestframework>=3.3.2']

setup_kwargs = {
    'name': 'drf-lookup-fields',
    'version': '1.1.0',
    'description': 'A package supplying tools for custom foreign-key fields lookup in drf-created WebAPI.',
    'long_description': '# Lookup Fields\n\n![Test package](https://github.com/innovationinit/drf-lookup-fields/actions/workflows/test-package.yml/badge.svg?branch=main)\n[![Coverage Status](https://coveralls.io/repos/github/innovationinit/drf-lookup-fields/badge.svg)](https://coveralls.io/github/innovationinit/drf-lookup-fields)\n\nThis package provides utils for changing lookup fields in requests to Django Rest Framework endpoints. It supports choosing lookup field for identifier of endpoint and identifiers used in related fields of serializers.\n\n## Per field lookup usage\n\nData for field should be supplied in `Content-Lookup-Fields` HTTP header as base64 encoded json object where keys are serializer fields names (if nested use dotted notation) and values are intended lookup fields:\n```json\n{\n  "some_field": "uuid",\n  "other.nested.field": "pk",\n}\n```\n\n### Field\n```python\nfrom lookup_fields.fields import CustomizableLookupRelatedField\nfrom rest_framework import serializers\n\n\nclass Serializer(serializers.ModelSerializer):\n    field = CustomizableLookupRelatedField(\n        lookup_fields={\n            \'some_field\': serializers.CharField(),\n            \'uuid\': serializers.UUIDField(),\n            \'other_field\': serializers.IntegerField(),\n            \'another_field_with_custom_filter_lookup_suffix\': {\n                \'field\': serializers.CharField(max_length=20),\n                \'filter_lookup_suffix\': \'__icontains\',\n            }\n        },\n        no_pk_lookup=True,\n        default_lookup_field_name=\'uuid\',\n    )\n```\n\ndefault_lookup_field_name is \'pk\', integer pk field is always added to lookup_fields. If you don\'t want this behavior set no_pk_lookup to True. Even then you can still add pk to lookup_fields when instantiate field.\n\nYou can specify optional parameters to lookup_fields by passing a dict instead of Field instance. Field has to be passed under \'field\' key.\n\n### Serializer\n\nSerializers uses `CustomizableLookupRelatedField` as default related field.\n\n```python\nfrom lookup_fields.fields import CustomizableLookupRelatedField\nfrom rest_framework import serializers\n\n\nclass Serializer(CustomizableLookupRelatedField):\n    class Meta:\n        model = SomeModel\n        fields = (\n            \'relation\',\n        )\n        extra_kwargs = {\n            \'relation\': {\n                \'lookup_fields\': {\n                    \'some_field\': serializers.CharField(),\n                },\n            },\n        }\n```\n\n## View mixin usage\nName of lookup field should be supplied in `Lookup-Field` HTTP header.\n\n### View\n```python\nfrom lookup_fields.views import CustomizableLookupFieldMixin\nfrom rest_framework.generics import GenericAPIView\n\n\nclass View(CustomizableLookupFieldMixin, GenericAPIView):\n    ALLOWED_LOOKUP_FIELDS = [\n        \'pk\',\n        \'uuid\',\n        \'some_field\',\n        WithPermissions(\n            \'other_field\',\n            permissions=[\'some_app.can_use_other_field_as_lookup_field\'],\n        )\n    ]\n    LOOKUP_FIELD_VALIDATORS = {\n        \'some_field\': [validator],\n    }\n```\n\n## License\nThe Django Rest Framework Lookup Fields package is licensed under the [FreeBSD\nLicense](https://opensource.org/licenses/BSD-2-Clause).\n',
    'author': 'IIIT',
    'author_email': 'github@iiit.pl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/innovationinit/drf-lookup-fields',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
